/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the bit library.                                 *
 *                                                                         *
 *   The bit library is free software; you can redistribute it and/or      *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The bit library is distributed in the hope that it will be useful,    *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#ifndef BITBUFFER_H
#define BITBUFFER_H

#include <bit/enums.h>
#include <bit/data.h>
#include <bit/pointer.h>

#include <bit/bit-config.h>

#ifdef BIT_HAVE_PTHREAD_H
#include <pthread.h>
#endif

namespace bit
  {

  /**
   * This class adds an interface for managing a packed bit (or octet) buffer.
   * to the base Data class which is octet oriented. Although this class
   * manages bit oriented buffers, it attempts to optimize for octet (byte)
   * based buffers.
   *
   * The methods pack(), unpack() and clear() provide a means of extracting
   * and setting bit segments in the underlying buffer.
   *
   * @ingroup buffers
   * @author Rick L. Vinyard, Jr.
   */
  class Buffer: public Data
    {

    public:

      /**
       * Construct a buffer of a given size that is allowed to dynamically grown as
       * pack/unpack requests are made.
       */
      Buffer(size_t initial_size=0, size_t sizemax=0) throw (std::bad_alloc);

      /** Constructor initializes buffer with a copy of specific data */
      Buffer(void* data, size_t size, size_t sizemax=0, DataMode mode=COPY) throw (std::bad_alloc);

      /** Constructor initializes buffer to use the specified underlying data chunk */
      Buffer( const Data& other );

      /** Copy constructor */
      Buffer( const Buffer& other );

      ~Buffer();

      /**
       * Transfer data from the buffer area into the low order bits of a memory location.
       * @param mem The memory location to transfer buffer data into
       * @param mem_octets The size of the memory location in octets
       * @param mem_loc The memory location relative position (END or START)
       * @param buf_offset_bits The offset from the beginning of the buffer in bits to begin extraction from
       * @param extract_bits The size of the data to extract in bits
       * @return true if transfer accomplished, false otherwise
       *
       * Bit Buffer
       * ==================================================
       * |000000000011111111112222222222333333333344444444|
       * |012345678901234567890123456789012345678901234567|
       * ==================================================
       *                ^ buf_offset_bits = 14
       *                ^ extract_bits = 20 ^
       *
       * Memory Buffer
       * ==================================================
       * |000000000011111111112222222222333333333344444444|
       * |012345678901234567890123456789012345678901234567|
       * ==================================================
       * ^ mem = start of memory address
       * ^ mem_octets = 6                                ^
       *                             ^ Resulting data    ^ when mem_loc = END
       * ^ Resulting data    ^                             when mem_loc = START
       */
      bool unpack(void* mem, size_t mem_octets, Location mem_loc, size_t buf_offset_bits, size_t extract_bits);

      Data unpack( size_t buf_offset_bits, size_t extract_bits, Location mem_loc=END );

      template <typename T>
      bool unpack(T& val, Location mem_loc, size_t buf_offset_bits, size_t extract_bits)
      {
        return unpack(&val, sizeof(T), mem_loc, buf_offset_bits, extract_bits);
      }

      /**
       * Transfer data from the low order bits of a memory location into the low order bits of a segment of the buffer area.
       * @param mem The memory location to transfer from
       * @param mem_octets The size of the memory location in octets
       * @param buf_offset The offset from the start of the data buffer in bits
       * @param buf_tgtsize The size of the target area in the buffer in bits
       * @param n The size of the data to extract in bits
       * @return true if transfer accomplished, false otherwise.
       * If @param offset + @param destsize are greater than the size of the buffer and the buffer
       * cannot be expanded to accommodate the request zero is returned.
       *
       * Memory Buffer
       * ==================================================
       * |000000000011111111112222222222333333333344444444|
       * |012345678901234567890123456789012345678901234567|
       * ==================================================
       * ^ mem = start of memory address
       * ^ mem_octets = 6                                ^
       *                                        ^ n = 10 ^ when mem_loc = END
       *                                        ^ cpy out^ when mem_loc = END
       * ^ n = 10 ^                                        when mem_loc = START
       * ^ cpy out^                                        when mem_loc = START
       *
       * Bit Buffer
       * ==================================================
       * |000000000011111111112222222222333333333344444444|
       * |012345678901234567890123456789012345678901234567|
       * ==================================================
       *                ^ buf_offset_bits = 14
       *                ^ buf_tgtsize  = 20 ^
       *                          ^ n = 10  ^
       *                          ^ cpy in  ^
       */
      bool pack(const void* mem, size_t mem_octets, Location mem_loc, size_t buf_offset, size_t buf_tgtsize, size_t n);

      bool pack( Data d, size_t buf_offset, size_t buf_tgtsize, size_t n, Location mem_loc=END );

      template <typename T>
      bool pack(const T& val, Location mem_loc, size_t offset, size_t destsize, size_t n)
      {
        bool b;
        b = pack(&val, sizeof(T), mem_loc, offset, destsize, n);
        return b;
      }

      bool clear_bits(size_t offset, size_t bits);

      #ifdef BIT_HAVE_PTHREAD_H
        bool read_lock();
        bool try_read_lock();
        bool write_lock();
        bool try_write_lock();
        bool unlock();
      #endif

      protected:

      #ifdef BIT_HAVE_PTHREAD_H
        struct RWLock {
          typedef BitPointer<RWLock> pointer;
          RWLock() {
            pthread_rwlock_init( &rwlock, NULL );
          }

          static pointer create() { return pointer( new RWLock() ); }

          ~RWLock() {
            pthread_rwlock_destroy( &rwlock );
          }

          pthread_rwlock_t rwlock;
        };

        RWLock::pointer m_rwlock;
      #endif

    };

}

#endif
