/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the bit library.                                 *
 *                                                                         *
 *   The bit library is free software; you can redistribute it and/or      *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The bit library is distributed in the hope that it will be useful,    *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include "data.h"

#include <glib.h>
#include <cstring>

namespace bit {

  Data::Data( const void* d, size_t s, DataMode mode ) throw ( std::bad_alloc )
  {
    void * ptr;

    m_data_storage = DataStorage::pointer( new DataStorage() );

    switch ( mode ) {
      case COPY:
        m_data_storage->manage_data = true;
        ptr = g_malloc( s );
        if ( !ptr )
          throw std::bad_alloc();
        m_data_storage->size = s;
        m_data_storage->data = ( uint8_t* ) ptr;
        ::memcpy( m_data_storage->data, d, s );
        break;
      case MANAGED:
        m_data_storage->manage_data = true;
        m_data_storage->size = s;
        m_data_storage->data = ( uint8_t* ) d;
        break;
      case UNMANAGED:
        m_data_storage->manage_data = false;
        m_data_storage->size = s;
        m_data_storage->data = ( uint8_t* ) d;
        m_data_storage->max_size = s;
        break;
    }

  }

  Data::Data( size_t s ) throw ( std::bad_alloc )
  {
    m_data_storage = DataStorage::pointer( new DataStorage() );

    m_data_storage->manage_data = true;

    if ( s != 0 )
    {
      void * ptr = ::calloc( s, 1 );
      if ( !ptr )
        throw std::bad_alloc();
      m_data_storage->size = s;
      m_data_storage->data = ( uint8_t* ) ptr;
    }
  }

  Data::Data( const Data& other ): m_data_storage( other.m_data_storage )
  {
  }

  Data::~Data()
  {
  }

  uint8_t* Data::data()
  {
    if ( ! m_data_storage ) return NULL;
    return m_data_storage->data;
  }

  const uint8_t* Data::data() const
  {
    if ( ! m_data_storage ) return NULL;
    return m_data_storage->data;
  }

  bool Data::set_data( const void* d, size_t s, DataMode mode ) throw ( std::bad_alloc )
  {
    uint8_t* ptr;

    // We don't deal with anything that overlaps the allocation
    if ( m_data_storage->data >= d && m_data_storage->data < (uint8_t*)d+s )
      return false;

    // if either the new data pointer is null or the size is zero, we have a request to clear the buffer
    if ( d == NULL || s == 0 ) {
      this->clear();
      return true;
    }

    // if our data size exceeds the max size then we limit the data size to the max size
    if ( m_data_storage->max_size)
      s = std::min( m_data_storage->max_size, s );

    switch ( mode ) {
      case COPY:

        // if we had external data, set the data pointer and size to null and 0
        // so that it will appear as if nothing has yet been allocated
        if ( ! m_data_storage->manage_data ) {
          m_data_storage->data = NULL;
          m_data_storage->size = 0;
        }

        // if it wasn't managed before, it's managed now
        m_data_storage->manage_data = true;

        // if we've reached this point we have a valid chunk of data to copy from.

        // We can always do a realloc, since a realloc with a null data buffer is equivalent to malloc.
        ptr = (uint8_t*) ::realloc( m_data_storage->data, s );
        if ( ptr == NULL )
          return false;
        m_data_storage->data = ptr;
        m_data_storage->size = s;
        ::memcpy( m_data_storage->data, d, s );
        break;
      case MANAGED:
        // if we had an allocation before, we need to free the memory
        if ( m_data_storage->data && m_data_storage->manage_data && m_data_storage->data != d )
        ::free( m_data_storage->data );

        m_data_storage->manage_data = true;
        m_data_storage->size = s;
        m_data_storage->data = ( uint8_t* ) d;
        break;
      case UNMANAGED:
        // if we had an allocation before, we need to free the memory
        if ( m_data_storage->data && m_data_storage->manage_data )
        ::free( m_data_storage->data );
        m_data_storage->manage_data = false;
        m_data_storage->size = s;
        m_data_storage->data = ( uint8_t* ) d;
        break;
    }

    return true;
  }

  size_t Data::size() const
  {
    return m_data_storage->size;
  }

  size_t Data::max_size( ) const
  {
    return m_data_storage->max_size;
  }

  void Data::set_max_size( size_t ms )
  {
    m_data_storage->max_size = ms;
  }

  bool Data::resize( size_t s ) throw ( std::bad_alloc )
  {
    uint8_t* ptr;

    // nothing to do if the size is already the right size, so
    // we're successful
    if ( m_data_storage->size == s )
      return true;

    // if we have external data, we have to fail
    if ( ! m_data_storage->manage_data )
      return false;

    // if the request is to resize to 0, that's really a request to clear
    if ( s == 0 ) {
      this->clear();
      return true;
    }

    // if our resize exceeds the max size then we fail
    if ( m_data_storage->max_size)
      return false;

    // At this point we really have to do some work
    // Now, realloc memory or if m_data_storage->data is NULL realloc will work like a malloc for us
    ptr = (uint8_t*) ::realloc( m_data_storage->data, s );

    // if the realloc failed we keep the memory the same and return false
    if ( ptr == NULL )
      return false;

    m_data_storage->data = ptr;

    // zero out newly allocated memory
    if ( s > m_data_storage->size )
      ::memset( m_data_storage->data + m_data_storage->size, 0x00, s - m_data_storage->size );

    m_data_storage->size = s;

    return true;
  }

  Data Data::clone() const
  {
    Data d( m_data_storage->data, m_data_storage->size, COPY );
    d.m_data_storage->max_size = m_data_storage->max_size;
    return d;
  }

  Data::operator bool()
  {
    return m_data_storage->data != NULL;
  }

  Data::operator bool() const
  {
    return m_data_storage->data != NULL;
  }

  Data::operator uint8_t*()
  {
    return m_data_storage->data;
  }

  Data::operator const uint8_t*() const
  {
    return m_data_storage->data;
  }

  std::string Data::encoded( Encoding en ) const
  {
    switch (en) {
    case ENCODING_NONE:
      if ( m_data_storage )
        return std::string( (const char*)(m_data_storage->data), m_data_storage->size );
      else
        return std::string();
    case ENCODING_TYPE:
      // TODO Add a type parameter and return a proper string
    case ENCODING_BASE64:
      return base64();
    case ENCODING_HEX:
      return hex();
    case ENCODING_OCT:
      return oct();
    case ENCODING_DEC:
      return dec();
    }

    // TODO throw a specific exception
    throw; // we shouldn't have reached here
  }
  
  std::string Data::base64() const
  {
    gchar* encoded;

    if ( ! m_data_storage ) return std::string();
    
    encoded = g_base64_encode((const guchar*)(m_data_storage->data), m_data_storage->size);

    std::string result(encoded);

    g_free(encoded);

    return result;
  }
  
  std::string Data::hex( std::string separator ) const
  {
    std::string s="0x";
    size_t i;
    bool have_separator = separator.length() > 0;
    uint8_t elem;
    char c[] = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F' };

    if ( ! m_data_storage ) return std::string();
    
    for ( i = 0; i < m_data_storage->size; i++ ) {
      if ( have_separator && i != 0 ) s += separator;
      elem = m_data_storage->data[i];
      s += c[ elem >> 4 ];
      s += c[ elem & 0x0F ];
    }
    return s;
  }

  std::string Data::oct( std::string separator ) const
  {
    std::string s="0";
    size_t i;
    bool have_separator = separator.length() > 0;
    uint8_t elem;
    char c[] = { '0', '1', '2', '3', '4', '5', '6', '7' };

    if ( ! m_data_storage ) return std::string();
    
    for ( i = 0; i < m_data_storage->size; i++ ) {
      if ( have_separator && i != 0 ) s += separator;
      elem = m_data_storage->data[i];
      s += c[ elem >> 6 ];
      s += c[ elem >> 3 & 007 ];
      s += c[ elem & 007 ];
    }
    return s;
  }

  std::string Data::dec( std::string separator ) const
  {
    std::string s;
    size_t i;
    bool have_separator = separator.length() > 0;
    uint8_t elem;
    char c[] = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9' };

    if ( ! m_data_storage ) return std::string();
    
    for ( i = 0; i < m_data_storage->size; i++ ) {
      if ( have_separator && i != 0 ) s += separator;
      elem = m_data_storage->data[i];
      if ( elem > 99 ) s += c[ elem / 100 ];
      if ( elem > 9 ) s += c[ elem % 100 / 10 ];
      s += c[ elem % 10 ];
    }
    return s;
  }

  bool Data::set_data(const std::string & encoded, Encoding en)
  {
    switch (en)
    {
      case ENCODING_BASE64:
        return this->set_base64_data(encoded);
      case ENCODING_HEX:
        return this->set_hex_data(encoded);
    }
    return false;
  }

  bool Data::set_base64_data(const std::string & encoded)
  {
    guchar* decoded;
    gsize decoded_size;

    decoded = g_base64_decode(encoded.c_str(), &decoded_size);

    this->set_data( decoded, decoded_size );
  }

  bool Data::set_hex_data(const std::string & encoded)
  {
    // i_e = index_encoded, i_d = index_decoded, i_dmax = index_decoded_max
    size_t i_e = 0, i_d = 0, i_dmax = encoded.length()/2;
    uint8_t decoded[i_dmax];
    bool have_upper=false;

      ::memset( decoded, 0x00, i_dmax );
    
    while ( i_e < encoded.size() && isspace(encoded[i_e]) ) i_e++; // skip all leading whitespace
    // check to see if we have a hex value of the form xnnnnn and increment i_e if we do
    if ( i_e < encoded.size() && tolower(encoded[i_e]) == 'x' ) i_e++;
    // check to see if we have a hex value of the form 0xnnnnn and increment i_e if we do
    if ( i_e+1 < encoded.size() && encoded[i_e] == '0' && tolower(encoded[i_e+1]) == 'x' ) i_e+=2;

    // now it's time to process the data
    while( i_e < encoded.size() )
    {
      // skip anything that isn't a hex digit and set next character to read into lower nibble
      while ( i_e < encoded.size() && !isxdigit(encoded[i_e]) ) {
        i_e++;
        if ( have_upper ) i_d++;
        have_upper = false;
      }

      // we either stopped because we're at the end of the string, or we have a hex digit
      if ( i_e == encoded.size() ) break;

      if (have_upper)
      {
        // we have the upper nibble in the lower nibble
        // so let's move it to the upper nibble
        decoded[i_d] <<= 4;
        // and now bring in the lower nibble
        if (isdigit(encoded[i_e]))
          decoded[i_d] |= encoded[i_e] - '0';
        else
          decoded[i_d] |= tolower(encoded[i_e]) - 'a' + 10;
        have_upper = false;
        i_d++;
      }
      else
      {
        if (isdigit(encoded[i_e]))
          decoded[i_d] |= encoded[i_e] - '0';
        else
          decoded[i_d] |= tolower(encoded[i_e]) - 'a' + 10;
        have_upper = true;
      }

      i_e++;
    }

    if ( have_upper )
    {
      decoded[i_d] <<= 4;
      i_d++;
    }

    return this->set_data(decoded, i_d);
  }

  void Data::clear()
  {
      g_free( m_data_storage->data );
      m_data_storage->data = NULL;
      m_data_storage->size = 0;
      m_data_storage->manage_data = true;
  };

  bool Data::operator <( const Data& other ) const
  {
    return this->compare(other) < 0;
  }

  bool Data::operator <=( const Data& other ) const
  {
    return this->compare(other) <= 0;
  }

  bool Data::operator ==( const Data& other ) const
  {
    return this->compare(other) == 0;
  }

  bool Data::operator !=( const Data& other ) const
  {
    return this->compare(other) != 0;
  }

  bool Data::operator >=( const Data& other ) const
  {
    return this->compare(other) >= 0;
  }

  bool Data::operator >( const Data& other ) const
  {
    return this->compare(other) > 0;
  }

  int Data::compare( const Data& other ) const
  {
    if ( m_data_storage->size == 0 )
    {
      if ( other.m_data_storage->size == 0 ) return 0;
      return -1;
    }

    if ( other.m_data_storage->size == 0 ) return 1;

    // but if we made it here, both have storage
    size_t comp_size = std::min(m_data_storage->size, other.m_data_storage->size );

    int result = ::memcmp( m_data_storage->data, other.m_data_storage->data, comp_size );

    if ( result != 0 )
      return result;
    if ( m_data_storage->size < other.m_data_storage->size )
      return -1;
    if ( m_data_storage->size > other.m_data_storage->size )
      return 1;
    return 0;
  }

}

