/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the bit library.                                 *
 *                                                                         *
 *   The bit library is free software; you can redistribute it and/or      *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The bit library is distributed in the hope that it will be useful,    *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#ifndef BIT_FIELD_H
#define BIT_FIELD_H

#include <string>
#include <vector>

#include <libxml++/libxml++.h>

#include <bit/fieldbase.h>

namespace bit
{

  /**
   * A Field identifies a starting point and a length.
   *
   * This class extends FieldBase by including a name and a description.
   *
   * @ingroup records
   * @author Rick L. Vinyard, Jr.
   */
  class Field: public FieldBase
  {
    public:

      typedef BitPointer<Field> pointer;

      Field ( std::string name="",
              size_t length=0,
              size_t length_units=BITS,
              int offset=0,
              size_t offset_units=BITS,
              std::string descr=""
            );

      static Field::pointer create ( std::string name="",
                                     size_t length=0,
                                     size_t length_units=BITS,
                                     int offset=0,
                                     size_t offset_units=BITS,
                                     std::string descr=""
                                   );

      virtual ~Field();

//       size_t read(xmlpp::Element& xml_element, size_t last_position);

      /** @return the length of this object in it's length units */
      virtual size_t length() const;

      virtual size_t length ( size_t units ) const;

      virtual void set_length ( size_t length );

      virtual size_t length_units() const;

      virtual void set_length_units ( size_t units );

      /** @return the offset of this object in the offset units of this object */
      virtual int offset() const;

      virtual int offset ( size_t units ) const;

      virtual void set_offset ( int offset );

      virtual size_t offset_units() const;

      virtual void set_offset_units ( size_t units );

      /**
       * Returns the string name associated with this object
       *
       * This pure virtual method is reimplemented by children to
       * provide the proper notation, such as the array notation of
       * elements, by children.
       */
      virtual std::string name() const;

      virtual std::string name ( int depth ) const;

      virtual void set_name ( const std::string& name );

      /**
       * Returns the string description of this object.
       */
      virtual std::string description() const;

      virtual void set_description ( const std::string& description );

      /** Returns a FieldType object that describes how the field data should be interpreted. */
      virtual const FieldType& type() const;

      /** Sets the FieldType object describing how the field data should be interpreted. */
      virtual void set_type( const FieldType& );

      /**
       * Return the data encoding type
       * 
       * If encoding type of this field is set to ENCODING_NONE and this field has
       * a parent, then the parent's encoding type is returned.
       */
      virtual Encoding encoding() const;

      /** Set the data encoding type */
      virtual void set_encoding( Encoding en );
      
      virtual const std::string& xml();

      virtual FieldBase::pointer clone();

      virtual Container container_type() { return FIELD; }

    protected:
      std::string m_name;
      size_t m_length;
      size_t m_length_units;
      int m_offset;
      size_t m_offset_units;
      std::string m_description;
      FieldType m_type;
      Encoding m_encoding;
  };

}

#endif
