/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the bit library.                                 *
 *                                                                         *
 *   The bit library is free software; you can redistribute it and/or      *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The bit library is distributed in the hope that it will be useful,    *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include "fieldbuffer.h"

#include "fieldvector.h"
#include "recordbuffer.h"
#include "utility.h"

#include <iostream>
#include <byteswap.h>
#include <cstring>

#include <algorithm>

namespace bit
{

  FieldBuffer::FieldBuffer() : m_buffer ( NULL )
  {}

  FieldBuffer::FieldBuffer ( RecordBuffer& b, FieldBase::pointer f ) : m_buffer ( &b ), m_field ( f )
  { }

  FieldBuffer::~FieldBuffer()
  { }

  void FieldBuffer::set_record_field ( RecordBuffer& b, FieldBase::pointer f )
  {
    m_buffer = &b;
    m_field = f;
  }

  void FieldBuffer::set_record_field ( RecordBuffer* b, FieldBase::pointer f )
  {
    m_buffer = b;
    m_field = f;
  }

  std::string FieldBuffer::xml( const std::string& padding )
  {
    std::ostringstream sout;

    sout << padding << "<data name=\"" << m_field->name();

    if ( m_field->container_type() == FIELD_ELEMENT )
      sout << "[" << bit_dynamic_pointer_cast<FieldVector::Element> ( m_field )->index() << "]";

    sout << "\">";

    switch ( m_field->type().type() )
    {
      default:
        sout << this->data().encoded(m_field->encoding());
    }

    sout << "</data>";

    return sout.str();
  }

  std::string FieldBuffer::value( )
  {
    std::string s;
    this->unpack(s);
    return s;
  }

  FieldBase::pointer FieldBuffer::field()
  {
    return m_field;
  }

  RecordBuffer* FieldBuffer::buffer()
  {
    return m_buffer;
  }

  FieldBuffer FieldBuffer::field ( size_t index ) throw ( exception::invalid_index )
  {
    if ( ! m_buffer )
      throw exception::no_recordbuffer();

    if ( m_field )
      return FieldBuffer ( *m_buffer, m_field->subfield ( index ) );
    else
      throw exception::invalid_index();
  }

  FieldBuffer FieldBuffer::field ( const std::string& index ) throw ( exception::invalid_index )
  {
    if ( ! m_buffer )
      throw exception::no_recordbuffer();

    if ( m_field )
      return FieldBuffer ( *m_buffer, m_field->subfield ( index ) );
    else
      throw exception::invalid_index();
  }

  FieldBuffer FieldBuffer::operator[] ( size_t index ) throw ( exception::invalid_index )
  {
    return this->field ( index );
  }

  FieldBuffer FieldBuffer::operator[] ( const std::string& index ) throw ( exception::invalid_index )
  {
    return this->field ( index );
  }

  FieldBuffer FieldBuffer::operator[] ( const char* index ) throw ( exception::invalid_index )
  {
    return this->field ( index );
  }

  Data FieldBuffer::data( )
  {
    Data d = Data ( m_field->length ( OCTETS ) );
    this->unpack ( d.data(), m_field->length ( OCTETS ) );
    return d;
  }

  bool FieldBuffer::unpack ( void * mem, size_t mem_octets )
  {
    if ( ! m_buffer || ! m_field )
      throw exception::no_recordbuffer();

    return m_buffer->unpack ( mem, mem_octets, END, m_field->start ( BITS ), m_field->length ( BITS ) );
  }

  bool FieldBuffer::pack ( const void * mem, size_t mem_octets )
  {
    bool b;

    if ( ! m_buffer || ! m_field )
      throw exception::no_recordbuffer();

    b = m_buffer->pack ( mem, mem_octets, END, m_field->start ( BITS ), m_field->length ( BITS ), m_field->length ( BITS ) );
    if ( b )
      //     m_buffer->on_field_changed(m_field->full_name(m_field->depth()-1));
      m_buffer->on_field_changed ( m_field );
    return b;
  }

  bool FieldBuffer::pack ( const void * mem, size_t mem_octets, size_t n )
  {
    bool b;

    if ( ! m_buffer || ! m_field )
      throw exception::no_recordbuffer();

    b = m_buffer->pack ( mem, mem_octets, END, m_field->start ( BITS ), m_field->length ( BITS ), n );
    if ( b )
      //     m_buffer->on_field_changed(m_field->full_name(m_field->depth()-1));
      m_buffer->on_field_changed ( m_field );
    return b;
  }

  bool FieldBuffer::unpack_uint64 ( uint64_t& ui64 )
  {
    ui64=0x00;

    if ( ! m_buffer || ! m_field )
      throw exception::no_recordbuffer();

    // ensure that length is no more than 64
    size_t length = std::min ( ( size_t ) 64, m_field->length ( BITS ) );

//     std::cout << "Unpacking " << m_field->name() << " with byte order " << m_field->type().byte_order() << std::endl;

    if ( m_field->type().byte_order() == BYTEORDER_LITTLE_ENDIAN )
    {
      if ( ! m_buffer->unpack ( &ui64, sizeof ( uint64_t ), START, m_field->start ( BITS ), length ) )
        return false;
      // need to expand upper octet
      if ( length%8 != 0 )
      {
        uint8_t* upper_octet = ( uint8_t* ) ( &ui64 ) + length/8;
        *upper_octet >>= 8 - length%8;
      }
      ui64 = le_to_host ( ui64 );
    }
    else
    {
      if ( ! m_buffer->unpack ( &ui64, sizeof ( uint64_t ), END, m_field->start ( BITS ), length ) )
        return false;
      if ( m_field->type().byte_order() == BYTEORDER_NETWORK )
        ui64 = net_to_host ( ui64 );
      else if ( m_field->type().byte_order() == BYTEORDER_BIG_ENDIAN )
        ui64 = be_to_host ( ui64 );
    }
    return true;
  }

  bool FieldBuffer::unpack_int64 ( int64_t& i64 )
  {
    uint64_t ui64;

    if ( ! m_buffer || ! m_field )
      throw exception::no_recordbuffer();

    if ( ! this->unpack_uint64( ui64 ) )
      return false;

    // At this point we have the number in ui64 as an unsigned int

    // Let's see if we have a negative number
    uint64_t mask = 1 << (m_field->length(BITS) - 1);

    // No matter what we do we need to copy the bit pattern of ui64 into i64
      ::memcpy( &i64, &ui64, sizeof(ui64) );

    // We have a negative number if the highest bit is a one
    if ( ! (mask & ui64) )
      // but in this case it wasn't, so let's copy and return
      return true;

    // But, if we're here we have a negative number, so we need to do what is necessary
    // to fix it up
    switch ( m_field->type().sign() )
    {
      case SIGN_UNSIGNED:
        // It's possible we made it here on an unsigned type that had a high bit of a 1
        // Let's do nothing since ui64 is already in i64 and return as successful...
        // TODO Should an overflow exception be thrown if length == 64???
        return true;
      case SIGN_TWOS_COMPLEMENT:
        // In this case we need to make all the upper bits ones
        while (mask != 0x8000000000000000LL)
        {
          i64 |= mask;
          mask <<= 1;
        }
        i64 |= mask;
        return true;
      case SIGN_ONES_COMPLEMENT:
        // Here, we need to do the same, but also add one at the end
        // In this case we need to make all the upper bits ones
        while (mask != 0x8000000000000000LL)
        {
          i64 |= mask;
          mask <<= 1;
        }
        i64 |= mask;
        i64 += 1;
        return true;
      case SIGN_BIT:
        // And, if we made it here, we need to remove the sign bit and return the magnitude * -1
        i64 &= ~mask;
        i64 *= -1;
        return true;
    }

    return false;
  }

  bool FieldBuffer::pack_uint64 ( uint64_t ui64 )
  {
    bool b;

    if ( ! m_buffer || ! m_field )
      throw exception::no_recordbuffer();

    // ensure that length is no more than 64
    size_t length = std::min ( ( size_t ) 64, m_field->length ( BITS ) );

    if ( m_field->type().byte_order() == BYTEORDER_LITTLE_ENDIAN )
    {
      ui64 = host_to_le ( ui64 );
      // need to compress upper octet
      if ( length%8 != 0 )
      {
        uint8_t* upper_octet = ( uint8_t* ) ( &ui64 ) + length/8;
        *upper_octet <<= 8 - length%8;
      }
      b = m_buffer->pack ( &ui64, sizeof ( uint64_t ), START, m_field->start ( BITS ), length, length );
    }
    else
    {
      if ( m_field->type().byte_order() == BYTEORDER_NETWORK )
        ui64 = host_to_net ( ui64 );
      else if ( m_field->type().byte_order() == BYTEORDER_BIG_ENDIAN )
        ui64 = host_to_be ( ui64 );
      b = m_buffer->pack ( &ui64, sizeof ( uint64_t ), END, m_field->start ( BITS ), length, length );
    }

    if ( b )
      //     m_buffer->on_field_changed(m_field->full_name(m_field->depth()-1));
      m_buffer->on_field_changed ( m_field );
    return b;
  }

  bool FieldBuffer::unpack_float ( float& f )
  {
    bool b;

    if ( ! m_buffer || ! m_field )
      throw exception::no_recordbuffer();

    b = this->unpack( &f, sizeof ( float ) );
    switch ( m_field->type().byte_order() )
    {
      case BYTEORDER_HOST:
        break;
      case BYTEORDER_NETWORK:
        f = net_to_host ( f );
        break;
      case BYTEORDER_BIG_ENDIAN:
        f = be_to_host ( f );
        break;
      case BYTEORDER_LITTLE_ENDIAN:
        f = le_to_host ( f );
        break;
    }
    return b;
  }

  bool FieldBuffer::unpack_double ( double& d )
  {
    bool b;

    if ( ! m_buffer || ! m_field )
      throw exception::no_recordbuffer();

    b = this->unpack( &d, sizeof ( double ) );
    switch ( m_field->type().byte_order() )
    {
      case BYTEORDER_HOST:
        break;
      case BYTEORDER_NETWORK:
        d = net_to_host ( d );
        break;
      case BYTEORDER_BIG_ENDIAN:
        d = be_to_host ( d );
        break;
      case BYTEORDER_LITTLE_ENDIAN:
        d = le_to_host ( d );
        break;
    }
    return b;
  }

  bool FieldBuffer::unpack_string( std::string& s )
  {
    if ( ! m_buffer || ! m_field )
      throw exception::no_recordbuffer();

    size_t octets = m_field->length(OCTETS);
    char* buf = new char[ octets+1 ];
    bool b = this->unpack( buf, octets );
    if ( b )
    {
      buf[octets] = 0x00;
      s = std::string( buf );
      delete[] buf;
      return true;
    }

    return false;
  }

  bool FieldBuffer::pack_double ( double d )
  {
    if ( ! m_buffer || ! m_field )
      throw exception::no_recordbuffer();

    // TODO put in proper throw
    if ( m_field->length ( BITS ) != 64 ) throw;
    switch ( m_field->type().byte_order() )
    {
      case BYTEORDER_HOST:
        break;
      case BYTEORDER_NETWORK:
        d = host_to_net ( d );
        break;
      case BYTEORDER_BIG_ENDIAN:
        d = host_to_be ( d );
        break;
      case BYTEORDER_LITTLE_ENDIAN:
        d = host_to_le ( d );
        break;
    }
    return this->pack( &d, sizeof ( double ) );
  }

  bool FieldBuffer::pack_float ( float f )
  {
    if ( ! m_buffer || ! m_field )
      throw exception::no_recordbuffer();

    // TODO put in proper throw
    if ( m_field->length ( BITS ) != 32 ) throw;
    switch ( m_field->type().byte_order() )
    {
      case BYTEORDER_HOST:
        break;
      case BYTEORDER_NETWORK:
        f = host_to_net ( f );
        break;
      case BYTEORDER_BIG_ENDIAN:
        f = host_to_be ( f );
        break;
      case BYTEORDER_LITTLE_ENDIAN:
        f = host_to_le ( f );
        break;
    }
    return this->pack( &f, sizeof ( float ) );
  }

}

