/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the bit library.                                 *
 *                                                                         *
 *   The bit library is free software; you can redistribute it and/or      *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The bit library is distributed in the hope that it will be useful,    *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include "fieldtype.h"

#include <endian.h>

namespace bit
{

  const char* FieldType::m_type_string[] =
      {
          "None",
          "Integer",
          "Floating Point",
          "ASCII",
          "BCD (Binary Coded Decimal)",
          "UTF8",
      };

  const char* FieldType::m_byte_order_string[] =
      {
          "Network",
          "Big Endian",
          "Little Endian",
      };

  const char* FieldType::m_sign_string[] =
      {
          "Unsigned",
          "Two's Complement",
          "One's Complement",
          "Sign Bit",
      };

  const char* FieldType::type_string ( Type type )
  {
    return m_type_string[type];
  }

  const char* FieldType::byte_order_string ( ByteOrder byteorder )
  {
    return m_byte_order_string[byteorder];
  }

  const char* FieldType::sign_string ( Sign sign )
  {
    return m_sign_string[sign];
  }

  FieldType::FieldType ( Type type, ByteOrder byte_order, Sign sign ) :
      m_type ( type ),
      m_byte_order ( byte_order ),
      m_sign ( sign )
  {}


  FieldType::~FieldType()
  {}


  Type FieldType::type() const
  {
    return m_type;
  }

  void FieldType::set_type ( Type type )
  {
    m_type = type;
  }

  const char* FieldType::type_string() const
  {
    return type_string ( m_type );
  }

  ByteOrder FieldType::byte_order() const
  {
    return m_byte_order;
  }

  void FieldType::set_byte_order ( ByteOrder byte_order )
  {
    m_byte_order = byte_order;
  }

  const char* FieldType::byte_order_string() const
  {
    return byte_order_string ( m_byte_order );
  }

  Sign FieldType::sign() const
  {
    return m_sign;
  }

  void FieldType::set_sign ( Sign sign )
  {
    m_sign = sign;
  }

  const char* FieldType::sign_string() const
  {
    return sign_string ( m_sign );
  }

  bool FieldType::operator == ( const FieldType & other ) const
  {
    return ( m_type == other.m_type && m_byte_order == other.m_byte_order && m_sign == other.m_sign );
  }

  bool FieldType::operator != ( const FieldType & other ) const
  {
    return ( m_type != other.m_type || m_byte_order != other.m_byte_order || m_sign != other.m_sign );
  }

  std::string FieldType::xml( ) const
  {
    std::string sxml;

    sxml = "<type>";

    switch ( m_type )
    {
      case TYPE_NONE:
        sxml = std::string();
        return sxml;
      case TYPE_INTEGER:
        sxml += "<integer sign=\"";
        switch ( m_sign )
        {
          case SIGN_UNSIGNED:
            sxml += "unsigned\"";
            break;
          case SIGN_TWOS_COMPLEMENT:
            sxml += "twos_complement\"";
            break;
          case SIGN_ONES_COMPLEMENT:
            sxml += "ones_complement\"";
            break;
          case SIGN_BIT:
            sxml += "sign_bit\"";
            break;
        }
        break;
      case TYPE_FLOATING:
        sxml += "<floating";
        break;
      case TYPE_ASCII:
        sxml += "<ascii";
        break;
      case  TYPE_UTF8:
        sxml += "<utf8";
        break;
      case  TYPE_BCD:
        sxml += "<bcd";
        break;
    }

    if ( m_type == TYPE_INTEGER || m_type == TYPE_FLOATING )
    {
      sxml += " byte_order=\"";

      switch ( m_byte_order )
      {
        case BYTEORDER_HOST:
# if __BYTE_ORDER == __BIG_ENDIAN
          sxml += "big_endian\"";
# elif __BYTE_ORDER == __LITTLE_ENDIAN
          sxml += "little_endian\"";
# endif
          break;
        case BYTEORDER_NETWORK:
          sxml += "network\"";
          break;
        case BYTEORDER_BIG_ENDIAN:
          sxml += "big_endian\"";
          break;
        case BYTEORDER_LITTLE_ENDIAN:
          sxml += "little_endian\"";
          break;
      }
    }

    sxml += "/></type>";

    return sxml;
  }

}

