/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the bit library.                                 *
 *                                                                         *
 *   The bit library is free software; you can redistribute it and/or      *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The bit library is distributed in the hope that it will be useful,    *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#ifndef BITRECORD_H
#define BITRECORD_H

#include <string>

#include <libxml++/libxml++.h>

#include <bit/recordbase.h>
#include <bit/recordstorage.h>

/**
 * @defgroup records "Record Definitions"
 * Records and Fields are used to describe a parse tree to be applied to a linear data buffer.
 */

namespace bit
  {

    /**
     * bit::Record is a tree of bit::Field and bit::Record types sorted by
     * their starting bit positions.
     *
     * FOLDOC defines @b record as:
     * "An ordered set of fields, usually stored contiguously. ...
     * Some programming languages use the term to mean a type composed of fields of
     * several other types (C calls this a "struct")."
     *
     * In the context of the bit library, a record stores an ordered set of bit::Field types.
     *
     * The Record class extends the RecordBase by providing storage for Field objects as
     * well as providing the concept of a name and description strings.
     *
     * @ingroup records
     * @author Rick L. Vinyard, Jr.
     */
  class Record: public RecordBase
    {
    public:

      /** Convenience typedef to a smart pointer to bit::Record */
      typedef BitPointer<Record> pointer;

      /** Default constructor */
      Record( std::string name="",
              size_t length_units=BITS,
              int offset=0,
              size_t offset_units=BITS,
              std::string descr=""
            );

      /** Static method returns a smart pointer to a new Record object */
      static Record::pointer create( std::string name="",
                                     size_t length_units=BITS,
                                     int offset=0,
                                     size_t offset_units=BITS,
                                     std::string descr=""
                                   );

      /** Destructor */
      virtual ~Record();

      /** @return the length of this object in it's length units */
      virtual size_t length() const;

      virtual size_t length(size_t units) const;

      virtual const std::string& xml();

      virtual iterator begin();

      virtual iterator end();

      virtual size_t subfields();

      virtual FieldBase::pointer subfield(const std::string& s);

      virtual FieldBase::pointer subfield(size_t i);

      virtual void add_field(FieldBase::pointer field);

      virtual void remove_field(FieldBase::pointer field);

      virtual void remove_field(const std::string& name);

      virtual FieldBase::pointer clone();

      virtual Container container_type() { return RECORD; }

      protected:

      RecordStorage m_subfields;

      virtual FieldBase::pointer previous_field(FieldBase::pointer current_field) throw (exception::invalid_container_op);

      virtual FieldBase::pointer next_field(FieldBase::pointer current_field) throw (exception::invalid_container_op);
    };

}

#endif
