/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the bit library.                                 *
 *                                                                         *
 *   The bit library is free software; you can redistribute it and/or      *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The bit library is distributed in the hope that it will be useful,    *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include "recordbuffer.h"

#include "recordvector.h"

namespace bit
{

  RecordBuffer::RecordBuffer ( size_t initial_size, size_t sizemax ) throw ( std :: bad_alloc ) :
      Buffer ( initial_size, sizemax )
  {
    m_recordbuffer_storage = RecordBufferStorage::pointer ( new RecordBufferStorage() );
  }

  RecordBuffer::RecordBuffer ( RecordBase::pointer record, size_t initial_size, size_t sizemax ) throw ( std :: bad_alloc ) :
      Buffer ( initial_size, sizemax ),
      m_record ( record )
  {
    m_recordbuffer_storage = RecordBufferStorage::pointer ( new RecordBufferStorage() );
  }

  RecordBuffer::RecordBuffer ( void * data, size_t size, size_t sizemax, DataMode mode ) throw ( std :: bad_alloc ) :
      Buffer ( data, size, sizemax, mode )
  {
    m_recordbuffer_storage = RecordBufferStorage::pointer ( new RecordBufferStorage() );
  }

  RecordBuffer::RecordBuffer ( RecordBase::pointer record, void * data, size_t size, size_t sizemax, DataMode mode ) throw ( std :: bad_alloc ) :
      Buffer ( data, size, sizemax, mode ),
      m_record ( record )
  {
    m_recordbuffer_storage = RecordBufferStorage::pointer ( new RecordBufferStorage() );
  }

  RecordBuffer::RecordBuffer ( const Data & other ) :
      Buffer ( other )
  {
    m_recordbuffer_storage = RecordBufferStorage::pointer ( new RecordBufferStorage() );
  }

  RecordBuffer::RecordBuffer ( RecordBase::pointer record, const Data & other ) :
      Buffer ( other ),
      m_record ( record )
  {
    m_recordbuffer_storage = RecordBufferStorage::pointer ( new RecordBufferStorage() );
  }

  RecordBuffer::RecordBuffer ( const RecordBuffer & other ) :
      Buffer ( other ),
      m_record ( other.m_record ),
      m_recordbuffer_storage ( other.m_recordbuffer_storage )
  {}

  RecordBuffer::~RecordBuffer()
  {}

  std::string RecordBuffer::xml( const std::string& padding )
  {
    return this->xml ( m_record, padding );
  }

  std::string RecordBuffer::xml ( FieldBase::pointer f, const std::string& padding )
  {
    std::ostringstream sout;
    std::string child_padding = padding + "  ";

    if ( f->container_type() == FIELD || f->container_type() == FIELD_ELEMENT )
      return FieldBuffer ( *this, f ).xml(padding);

    #ifdef BIT_HAVE_PTHREAD_H
      this->read_lock();
    #endif

    if ( f->container_type() == RECORD || f->container_type() == RECORD_VECTOR || f->container_type() == RECORD_ELEMENT )
    {
      sout << padding << "<data_record name=\"" << f->name();
      sout << "\">\n";
    }

    for ( size_t i = 0; i < f->subfields(); i++ )
      sout << this->xml ( f->subfield (i), child_padding ) << "\n";

    if ( f->container_type() == RECORD || f->container_type() == RECORD_VECTOR || f->container_type() == RECORD_ELEMENT )
      sout << padding << "</data_record>";

    #ifdef BIT_HAVE_PTHREAD_H
      this->unlock();
    #endif

    return sout.str();
  }

  FieldBuffer RecordBuffer::field ( size_t field_id ) throw ( exception::no_record )
  {
    if ( m_record )
      return FieldBuffer ( *this, m_record->subfield ( field_id ) );
    else
      throw exception::no_record();
  }

  FieldBuffer RecordBuffer::field ( const std::string& field_id ) throw ( exception::no_record )
  {
    if ( m_record )
      return FieldBuffer ( *this, m_record->subfield ( field_id ) );
    else
      throw exception::no_record();
  }

  FieldBuffer RecordBuffer::field ( FieldBase::iterator iter )
  {
    return FieldBuffer ( *this, iter.pointer() );
  }

  FieldBuffer RecordBuffer::field ( FieldBase::pointer pointer )
  {
    return FieldBuffer ( *this, pointer );
  }

  FieldBuffer RecordBuffer::operator[] ( size_t field_id ) throw ( exception::no_record )
  {
    return this->field ( field_id );
  }

  FieldBuffer RecordBuffer::operator[] ( const std::string& field_id ) throw ( exception::no_record )
  {
    return this->field ( field_id );
  }

  FieldBuffer RecordBuffer::operator[] ( const char* field_id ) throw ( exception::no_record )
  {
    return this->field ( field_id );
  }

  FieldBuffer RecordBuffer::operator[] ( FieldBase::iterator iter )
  {
    return this->field ( iter );
  }

  FieldBuffer RecordBuffer::operator[] ( FieldBase::pointer pointer )
  {
    return this->field ( pointer );
  }

  void RecordBuffer::set_record ( RecordBase::pointer record ) throw ()
  {
    m_record = record;
    m_recordbuffer_storage->signal_record_changed.emit();
  }

  void RecordBuffer::clear_record() throw ()
  {
    m_record.reset();
    m_recordbuffer_storage->signal_record_changed.emit();
  }

  RecordBase::pointer RecordBuffer::record( )
  {
    return m_record;
  }

  sigc::signal< void, FieldBuffer > RecordBuffer::signal_field_changed( )
  {
    return m_recordbuffer_storage->signal_field_changed;
  }

  sigc::signal<void> RecordBuffer::signal_record_changed( )
  {
    return m_recordbuffer_storage->signal_record_changed;
  }

  void RecordBuffer::on_field_changed ( FieldBase::pointer field )
  {
    FieldBase::pointer current;

    if ( m_recordbuffer_storage->signal_field_changed_emitting )
    {
      m_recordbuffer_storage->signal_field_changed_queue.push ( current );
    }
    else
    {

      m_recordbuffer_storage->signal_field_changed_emitting = true;
      current = field;

      do
      {
        // At this point, size is zero if we are the one that grabbed the
        // signal field changed emitting, so we'll drop through this and
        // emit the signal
        if ( m_recordbuffer_storage->signal_field_changed_queue.size() > 0 )
        {
          current = m_recordbuffer_storage->signal_field_changed_queue.front();
          m_recordbuffer_storage->signal_field_changed_queue.pop();
        }
        m_recordbuffer_storage->signal_field_changed.emit ( FieldBuffer ( *this, current ) );

        // now, we'll check and see if the queue grew while we were emitting
      }
      while ( m_recordbuffer_storage->signal_field_changed_queue.size() > 0 );

      m_recordbuffer_storage->signal_field_changed_emitting = false;
    }

  }

}

