/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the bit library.                                 *
 *                                                                         *
 *   The bit library is free software; you can redistribute it and/or      *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The bit library is distributed in the hope that it will be useful,    *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#ifndef BITRECORDBUFFER_H
#define BITRECORDBUFFER_H

#include <string>
#include <queue>
#include <set>

#include <bit/buffer.h>
#include <bit/record.h>
#include <bit/fieldbuffer.h>
#include <bit/data.h>
#include <bit/except.h>

namespace bit
  {

  /**
   * Combines the concepts of a Record or data layout with actual data.
   *
   * @ingroup buffers
   * @author Rick L. Vinyard, Jr.
   */
  class RecordBuffer : public Buffer
    {
    public:
      /**
       * Construct a buffer of a given size that is allowed to dynamically grown as
       * pack/unpack requests are made.
       */
      RecordBuffer(size_t initial_size=0, size_t sizemax=0) throw (std::bad_alloc);

      /**
       * Construct a buffer of a given size that is allowed to dynamically grown as
       * pack/unpack requests are made.
       */
      RecordBuffer(RecordBase::pointer record, size_t initial_size=0, size_t sizemax=0) throw (std::bad_alloc);

      /** Constructor initializes buffer with a copy of specific data */
      RecordBuffer(void* data, size_t size, size_t sizemax=0, DataMode mode=COPY) throw (std::bad_alloc);

      /** Constructor initializes buffer with a copy of specific data */
      RecordBuffer(RecordBase::pointer record, void* data, size_t size, size_t sizemax=0, DataMode mode=COPY) throw (std::bad_alloc);

      /** Constructor initializes buffer to use the specified underlying data chunk */
      RecordBuffer( const Data& other );

      /** Constructor initializes buffer to use the specified underlying data chunk */
      RecordBuffer( RecordBase::pointer record, const Data& other );

      /** Copy constructor */
      RecordBuffer( const RecordBuffer& other );

      ~RecordBuffer();

      std::string xml(const std::string& padding=std::string());

      std::string xml(FieldBase::pointer i, const std::string& padding=std::string());

      FieldBuffer field(size_t) throw (exception::no_record);
      FieldBuffer field(const std::string&) throw (exception::no_record);
      FieldBuffer field(FieldBase::iterator);
      FieldBuffer field(FieldBase::pointer);

      FieldBuffer operator[](size_t) throw (exception::no_record);
      FieldBuffer operator[](const std::string&) throw (exception::no_record);
      FieldBuffer operator[](const char*) throw (exception::no_record);
      FieldBuffer operator[](FieldBase::iterator);
      FieldBuffer operator[](FieldBase::pointer);

      void set_record(RecordBase::pointer record) throw ();
      void clear_record() throw ();
      RecordBase::pointer record();

      sigc::signal<void, FieldBuffer> signal_field_changed();

      sigc::signal<void> signal_record_changed();

    protected:
      friend class FieldBuffer;

      /**
       * A pointer to the record associated with this buffer, or NULL
       * if there is no record currently associated.
       */
      RecordBase::pointer m_record;

      /**
       * Causes the internal signal to be emitted notifying connected slots
       * that the data at field i has been changed.
       *
       * This method provides proper calling of the emit signal even when
       * a callback triggers a change. This would result in a recursive signal
       * which sigc++ doesn't allow.
       */
      void on_field_changed(FieldBase::pointer field);

      struct RecordBufferStorage
      {
        typedef BitPointer<RecordBufferStorage> pointer;

        RecordBufferStorage(): signal_field_changed_emitting(false) { }

        sigc::signal<void> signal_record_changed;

        /**
         * The signal emitted when a field's value is changed.
         *
         * This member has been move to private, since derived children should not
         * directly emit the signal, but instead should call the notify_of_changes_to_fields
         * method.
         *
         * This method properly ensures that a signal is emitted for each changed fields,
         * even when a callback triggers an fields to change. This is necessary since
         * sigc++ does not allow recursive signals anymore.
         */
        sigc::signal<void, FieldBuffer> signal_field_changed;

        // Stores the fields that have changed and still need to have an emit signal
        //called. The fields are stored in the order they are changed.
        std::queue< FieldBase::pointer > signal_field_changed_queue;

        bool signal_field_changed_emitting;
      };

      RecordBufferStorage::pointer m_recordbuffer_storage;

    };

}

#endif
